#include <stdio.h>
#include <string.h>
#include <process.h>
#include <windows.h>

#if defined(STANDALONE)
  #include "StackWalker.cpp"
  #include "exc-abort.cpp"
#endif

/*
 * We can compile this as pure C (when 'STANDALONE' is not defined),
 * but lets test the C++ un-decorations performed by StackWalker.cpp.
 */
#ifdef __cplusplus
namespace foo {
#endif

void warning (void)
{
#ifndef _MSC_VER
  puts ("Debug-symbols not available in this non-MS compiler.");
#endif
}

/* This should be compiled with "cl -Od -Oi-" to prevent functions below
 * to be optimized into one function.
 */
void func3 (void)
{
  puts ("Calling abort().");
  abort();
}

void func2 (void)
{
  func3();
}

void func1 (void)
{
  func2();
}

/*

  Here is what you get from StackWalker.cpp:

  f:\vc_2019\VC\Projects\Aborter> test-cl.exe -abort
  Calling abort().
  Got SIGABRT. Backtrace:
    stackwalker.cpp (1041): StackWalker::ShowCallstack
    exc-abort.cpp (82): new_abort_handler
    winsig.c (586): raise
    abort.c (74): abort
    test.cpp (21): foo::func3
    test.cpp (26): foo::func2
    test.cpp (31): foo::func1
    test.cpp (73): foo::test_abort
    test.cpp (94): main
    crtexe.c (555): __tmainCRTStartup
    crtexe.c (371): mainCRTStartup
    7C817077 (kernel32): (filename not available): RegisterWaitForInputIdle

  Here is what you normally gets when you omit the call to LoadLibrary ("exc-abort.dll").
  First you'll get a assert-box with "Abort", "Retry" and "Ignore". Pressing "Retry" gives
  you this; from WinDbg (my JIT-debugger):

    MSVCR100D!_NMSG_WRITE(int rterrnum = 0n10)+0x7b
    MSVCR100D!abort(void)+0x1e
    test_cl!foo::func3(void)+0x17
    test_cl!foo::func2(void)+0x8
    test_cl!foo::func1(void)+0x8
    test_cl!foo::test_abort(void)+0x8
    test_cl!main(int argc = 0n2, char ** argv = 0x0038aa98)+0x27
    test_cl!__tmainCRTStartup(void)+0x1bf
    test_cl!mainCRTStartup(void)+0xf
    kernel32!BaseProcessStart+0x23

*/

int test_abort (void)
{
  warning();
  func1();
  return (0);
}

int test_except (void)
{
  char *p = NULL;

  warning();
  *p = 0;              /* Baboom! */
  return (0);
}

#ifdef __cplusplus
};

using namespace foo;
#endif


int main (int argc, const char **argv)
{
#if defined(STANDALONE)
  init_exc_abort();
#else
  LoadLibrary ("exc-abort.dll");
#endif

  if (argc > 1 && !_stricmp(argv[1], "-abort"))
     return test_abort();

  if (argc > 1 && !_stricmp(argv[1], "-except"))
     return test_except();

  printf ("Usage: %s [-abort | -except]\n", argv[0]);
  return (0);
}
